#######################################################################
# Copyright (c) 2023 Juan Hern�ndez, FlowJo, LLC
#
# License
# The software is distributed under the terms of the 
# Artistic License 2.0
# http://www.r-project.org/Licenses/Artistic-2.0
# 
# Disclaimer
# This software and documentation come with no warranties of any kind.
# This software is provided "as is" and any express or implied 
# warranties, including, but not limited to, the implied warranties of
# merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the  copyright holder be liable for any direct, 
# indirect, incidental, special, exemplary, or consequential damages
# (including but not limited to, procurement of substitute goods or 
# services; loss of use, data or profits; or business interruption)
# however caused and on any theory of liability, whether in contract,
# strict liability, or tort arising in any way out of the use of this 
# software.
######################################################################
## For debugging to see what FlowJo is passing to the script.
##
## cat("FJ-PAR-PREFIX                = FJ_PAR_PREFIX \n")
## cat("FJ-SCRIPT-DIR			     = FJ_SCRIPT_DIR \n")
## cat("FJ-OUTPUT-DIR			     = FJ_OUTPUT_DIR \n")
## cat("FJ-DATA-FILE-FIRST-PATH	     = FJ_DATA_FILE_FIRST_PATH \n")
## cat("FJ-DATA-FILE-SECOND-PATH     = FJ_DATA_FILE_SECOND_PATH \n")
## cat("FJ-CLUSTERS-OUTPUT-FILE      = FJ_CLUSTERS_OUTPUT_FILE \n")
## cat("FJ-CSV-OUTPUT-FILE           = FJ_CSV_OUTPUT_FILE \n")
## cat("FJ-DIFF-OUTPUT-FILE          = FJ_DIFF_OUTPUT_FILE \n")
## cat("FJ-COMPARATION-METHOD        = FJ_COMPARATION_METHOD \n")
## cat("FJ-PAR-K                   	 = FJ_PAR_K \n")
## cat("FJ-PAR-BINS                  = FJ_PAR_BINS \n")
## cat("FJ-PAR-EPS                   = FJ_PAR_EPS \n")
## cat("FJ-TSNE-PERPLEXITY           = FJ_TSNE_PERPLEXITY \n")
## cat("FJ-TSNE-ITERATIONS           = FJ_TSNE_ITERATIONS \n")
## cat("FJ-MIN-CLUSTER-SIZE          = FJ_MIN_CLUSTER_SIZE \n")

# Install function for packages
loadPackages <-function(...) {
  libs<-unlist(list(...))
  req<-unlist(lapply(libs,require,character.only=TRUE))
  need<-libs[req==FALSE]
  if(length(need)>0){
    install.packages(need, repos="https://cran.r-project.org")
    lapply(need,require,character.only=TRUE)
  }
}

loadBioc<-function(...) {
  libs<-unlist(list(...))
  req<-unlist(lapply(libs,require,character.only=TRUE))
  need<-libs[req==FALSE]
  if(length(need)>0){
    BiocManager::install(need)
    lapply(need,require,character.only=TRUE)
  }
}

loadPackages(c('tidyverse','ggplot2','cowplot','FNN','dbscan','fancycut','gridExtra','uwot',
               'RColorBrewer','dplyr','purrr','Rtsne','BiocManager'))

loadBioc(c('cytoMEM','flowCore','Biobase'))

setwd("FJ_SCRIPT_DIR")

# load T-REX function from local folder
source("R/TREX.R")
source("R/TREX_clusters.R")
source("R/themes_and_palettes.R")
source("R/TREX_helpers.R")

## For debugging purposes
sessionInfo()

## Returns a list of 3 elements
## X = a frame of data from a CSV file as exported by FlowJo
## cellIds = a matrix with CellIds (if available)
## eventNumbers = a matrix with EventNumberDP (if available)
##
## csvPath: path the the CSV file as exported from FlowJo (including header with both stain and parameter is recommended when exporting from FlowJo)
readFlowJoCSV <- function(csvPath) {
  EventNumberDP <- "EventNumberDP"
  CellIdDP <- "CellId"

  data <- read.csv(csvPath, check.names=FALSE)

  cellIdsColumn <- rep(NA, nrow(data))
  cellIds <- data.frame(CellIdDP = cellIdsColumn)
  names(cellIds) <- CellIdDP
  if (CellIdDP %in% names(data)) {
    cellIds <- data[, CellIdDP]
    data <- data[, -which(names(data) %in% c(CellIdDP))]
  }

  EventNumberColumn <- rep(NA, nrow(data))
  eventNumbers <- data.frame(EventNumberDP = EventNumberColumn)
  names(eventNumbers) <- EventNumberDP
  if (EventNumberDP %in% names(data)) {
    eventNumbers <- data[, EventNumberDP]
    data <- data[, -which(names(data) %in% c(EventNumberDP))]
  }

  # Trim the leading and tailing whitespaces
  colnames(data) <- gsub("^\\s+|\\s+$", "", colnames(data))
  list(X=data, cellIds=as.matrix(cellIds), eventNumbers=as.matrix(eventNumbers))
}

# setup sample and T-REX info
parPrefix = "FJ_PAR_PREFIX"
csvOutputFile = "FJ_CSV_OUTPUT_FILE"
clustersOutputFile = "FJ_CLUSTERS_OUTPUT_FILE"
diffOutputFile = "FJ_DIFF_OUTPUT_FILE"
minClusterSize = FJ_MIN_CLUSTER_SIZE
runID <- "FJ_RUNID"

first.dataset = 1
second.dataset = 2

# combine paired samples
first.dataset.read = readFlowJoCSV("FJ_DATA_FILE_FIRST_PATH")
first.dataset.data = first.dataset.read$X
orig.names <- colnames(first.dataset.data)
orig.ncols = ncol(first.dataset.data)
orig.filenames <- c("FJ_DATA_FILE_FIRST_PATH","FJ_DATA_FILE_SECOND_PATH")
first.dataset.data$file_ID = "FJ_FIRST_DATASET_NAME"
first.dataset.data$eventNumbers = first.dataset.read$eventNumbers
colnames(first.dataset.data)

second.dataset.read = readFlowJoCSV("FJ_DATA_FILE_SECOND_PATH")
second.dataset.data = second.dataset.read$X
second.dataset.data$file_ID = "FJ_SECOND_DATASET_NAME"
second.dataset.data$eventNumbers = second.dataset.read$eventNumbers
colnames(second.dataset.data)

data.filenames = c("FJ_FIRST_DATASET_NAME","FJ_SECOND_DATASET_NAME")

#```{r equal sampling}
# find number of data points in smaller sample
smaller_sample = nrow(first.dataset.data)
if (nrow(second.dataset.data) < smaller_sample) {
  smaller_sample <- nrow(second.dataset.data)
}

# set a seed to get a reproducible sample
set.seed(1)

# get an equal number of data points from both data sets
complete_sampled.data = as.data.frame(rbind(
  first.dataset.data[sample(nrow(first.dataset.data), smaller_sample), ],
  second.dataset.data[sample(nrow(second.dataset.data), smaller_sample), ]
))
sampled.data = complete_sampled.data;
sampled.data$eventNumbers = NULL;

#```{r choose markers, scale data}
# choose markers to make low dimensional embedding
chosen.data = sampled.data
chosen.data$file_ID = NULL
chosen.data$eventNumbers = NULL
# scale data
transformed.data = chosen.data
#```

if("FJ_COMPARATION_METHOD" == "UMAP"){
  library(uwot)
  # create UMAP (low dimensional embedding)
  myumap <- umap(transformed.data, ret_model = TRUE, verbose = TRUE)
  comparation.data = as.data.frame(myumap$embedding)
  colnames(comparation.data) <- c("UMAP1", "UMAP2")
  comparation1 = sym("UMAP1")
  comparation2 = sym("UMAP2")
}else{
  library(Rtsne)
  mytsne <- Rtsne(transformed.data, dims = 2, perplexity=FJ_TSNE_PERPLEXITY, verbose=TRUE, max_iter = FJ_TSNE_ITERATIONS)
  comparation.data = as.data.frame(mytsne$Y)
  row.names(comparation.data) <- rownames(transformed.data)
  colnames(comparation.data) <- c("tSNE1", "tSNE2")
  comparation1 = sym("tSNE1")
  comparation2 = sym("tSNE2")
}

#```{r create output folder}
output.folder = "FJ_OUTPUT_DIR"
dir.create(output.folder)
setwd(output.folder)

my.binned = TREX(cbind(comparation.data, file_ID = sampled.data$file_ID),
                 kvalue = FJ_PAR_K,
                 bins = c(FJ_PAR_BINS)
)

TREX_plot(
  my.binned,
  embed.type = "FJ_COMPARATION_METHOD",
  caption = "",
  export = TRUE
)

# The plot title will be incorrectly placed on the image below, but correctly placed on the exported png. The location is based on aspect ratio of the image, and can be adjusted with a title.height parameter if needed.
TREX_results(my.binned, export = TRUE, exportFileName = diffOutputFile)

my.clusters = TREX_cluster(
  binned.data = my.binned,
  db.eps = FJ_PAR_EPS,
  marker.data = sampled.data
)

print('Cluster : Count')
for(cluster in unique(my.clusters$cluster)){
  print(paste(cluster, ':', length(my.clusters$cluster[my.clusters$cluster == cluster])))
}
if(minClusterSize > 1) {
  maxclusters <- max(my.clusters$cluster)
  minclusters <- min(my.clusters$cluster)
  cluster <- minclusters
  while(cluster <= maxclusters){
    nclusters <- length(my.clusters$cluster[my.clusters$cluster == cluster])
    if(nclusters < minClusterSize && nclusters > 0){
      print(paste('removing cluster ',cluster,' with ', nclusters, ' events'))
      my.clusters$cluster[my.clusters$cluster == cluster] <- -1
      for(c in cluster+1:maxclusters){
        my.clusters$cluster[my.clusters$cluster == c] <- c - 1
      }
      cluster <- cluster - 1
    }
    cluster <- cluster + 1
  }
}
print('Cluster : Count')
for(cluster in unique(my.clusters$cluster)){
  print(paste(cluster, ':', length(my.clusters$cluster[my.clusters$cluster == cluster])))
}

TREX_cluster_plot(
  cluster.data = my.clusters,
  binned.data = my.binned,
  embed.type = "FJ_COMPARATION_METHOD",
  export = TRUE,
  run.ID = "FJ_RUNID"
)

TREX_cluster_results(my.clusters, export = TRUE)

# run MEM on DBSCAN clusters
MEM.input = my.clusters[, c(6:ncol(my.clusters))]
MEM.input <- MEM.input[order(MEM.input$cluster), ]

# JH MEM not works correctly if cluster number is <= 1
num_pops <- length(unique(MEM.input$cluster))
if(num_pops > 1){
  #help MEM(exp_data, transform = FALSE, cofactor = 1, choose.markers = FALSE,
  #  markers = "all", choose.ref = FALSE, zero.ref = FALSE, rename.markers = FALSE,
  #  new.marker.names = "none", file.is.clust = FALSE, add.fileID = FALSE,
  #  IQR.thresh = NULL, output.prescaled.MEM = FALSE, scale.matrix = "linear",
  #  scale.factor = 0)
  MEM.output = MEM(
    MEM.input,
    transform = FALSE,
    zero.ref = TRUE,
    scale.matrix = "arcsinh",
    scale.factor = 1
  )

  build_heatmaps(
    MEM.output,
    cluster.MEM = "none",
    cluster.medians = "none",
    display.thresh = 1,
    output.files = TRUE,
    labels = FALSE
  )
}

#move MEM files to output_file
mem_folder = paste0(output.folder, "/output files/")
my_files <- list.files(mem_folder)
file.copy(from = paste0(mem_folder, my_files),
          to = paste0(output.folder,"/", my_files))
file.remove(from = paste0(mem_folder, my_files))
unlink(mem_folder, recursive = TRUE)

#```{r count cells in each cluster}
TREX_counts(my.clusters, export = TRUE, exportFileName = clustersOutputFile)


comparation.bychannel <- as_tibble(comparation.data) %>%
  bind_cols(transformed.data)  %>%
  gather(channel, intensity, -!!comparation1, -!!comparation2) %>%
  mutate(across(channel,factor))%>%
  group_split(channel) %>%
  map(
    ~ggplot(., aes(x = !!comparation1, y = !!comparation2, col = intensity)) +
      geom_point(shape = ".", size = 8) +
      scale_color_gradientn(
        colours = colorRampPalette(rev(brewer.pal(n = 11, name = "Spectral")))(5)) +
      facet_grid(~ channel, labeller = function(x) label_value(x, multi_line = FALSE)) +
      coord_fixed() +
      theme_bw() +
      theme(
        strip.text.x = element_text(size = 20),
        legend.title = element_blank(),
        panel.grid = element_blank(),
        axis.title = element_blank(),
        axis.text = element_blank(),
        axis.ticks = element_blank()
      )
  ) %>%
  plot_grid(plotlist = ., align = 'hv', ncol = 8)

png(
  paste(strftime(Sys.time(),"%Y-%m-%d_%H-%M"),"_marker_heat.png"),
  height = 2000,
  width = 4000
)
print(comparation.bychannel)
dev.off()
# export new FCS files with sampled data, percent change, and tsne axes
to.export = cbind(complete_sampled.data, comparation.data, percent.change = my.binned$percent.change)

# add column for cluster number
to.export$cluster <- c(0)
bins.of.interest = unique(my.clusters$cuts)
temp.cuts = wafflecut(to.export$percent.change, levels(my.binned$cuts))
for(n in rownames(to.export[temp.cuts %in% bins.of.interest, ])){
  to.export[n, "cluster"] <- my.clusters[n, "cluster"]
}

# save the result to CSV
to.csv = cbind(complete_sampled.data$eventNumbers, comparation.data, my.binned$percent.change, to.export$cluster)
names(to.csv)[1] <- "EventNumberDP"
names(to.csv)[2] <- paste(parPrefix, names(comparation.data)[1], runID, sep = "_")
names(to.csv)[3] <- paste(parPrefix, names(comparation.data)[2], runID, sep = "_")
names(to.csv)[4] <- paste(parPrefix, "percent_change", runID, sep = "_")
names(to.csv)[5] <- paste(parPrefix, "cluster", runID, sep = "_")
to.csv <- to.csv[order(to.csv$EventNumberDP),]
write.csv(to.csv, file=csvOutputFile, row.names=FALSE, quote=FALSE)
#```
rm(list = ls())